@@ -1,30 +1,8 @@
 /*
- * Copyright (c) 2004-2007 Hannes Reinecke and Douglas Gilbert.
+ * Copyright (c) 2004-2015 Hannes Reinecke and Douglas Gilbert.
  * All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- *
+ * Use of this source code is governed by a BSD-style
+ * license that can be found in the BSD_LICENSE file.
  */
 
 #include <unistd.h>
@@ -41,6 +19,8 @@
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
 /* A utility program for the Linux OS SCSI subsystem.
  *
@@ -48,16 +28,16 @@
  *  to the 'SCSI Accessed Fault-Tolerant Enclosures' (SAF-TE) spec.
  */
 
-static char * version_str = "0.23 20071002";
+static const char * version_str = "0.26 20151219";
 
 
-#define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
+#define SENSE_BUFF_LEN 64       /* Arbitrary, could be larger */
 #define DEF_TIMEOUT 60000       /* 60,000 millisecs == 60 seconds */
 #define EBUFF_SZ 256
 
 #define RB_MODE_DESC 3
-#define RWB_MODE_DATA 2 
-#define RWB_MODE_VENDOR 1 
+#define RWB_MODE_DATA 2
+#define RWB_MODE_VENDOR 1
 #define RB_DESC_LEN 4
 
 #define SAFTE_CFG_FLAG_DOORLOCK 1
@@ -95,14 +75,14 @@ read_safte_configuration(int sg_fd, unsigned char *rb_buff,
     int res;
 
     if (rb_len < buf_capacity) {
-        fprintf(stderr,"SCSI BUFFER size too small (%d/%d bytes)\n",
-                rb_len, buf_capacity);
+        pr2serr("SCSI BUFFER size too small (%d/%d bytes)\n", rb_len,
+                buf_capacity);
         return SG_LIB_CAT_ILLEGAL_REQ;
     }
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=0 "
-                "to fetch configuration\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=0 to fetch "
+                "configuration\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 0, 0,
                             rb_buff, rb_len, 1, verbose);
     if (res && res != SG_LIB_CAT_RECOVERED)
@@ -112,7 +92,7 @@ read_safte_configuration(int sg_fd, unsigned char *rb_buff,
     safte_cfg.psupplies = rb_buff[1];
     safte_cfg.slots = rb_buff[2];
     safte_cfg.temps = rb_buff[4];
-    if (rb_buff[3]) 
+    if (rb_buff[3])
         safte_cfg.flags |= SAFTE_CFG_FLAG_DOORLOCK;
     if (rb_buff[5])
         safte_cfg.flags |= SAFTE_CFG_FLAG_ALARM;
@@ -153,8 +133,8 @@ do_safte_encl_status(int sg_fd, int do_hex, int do_raw, int verbose)
 
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=1 "
-                "to read enclosure status\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=1 to read "
+                "enclosure status\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 1, 0,
                             rb_buff, rb_len, 0, verbose);
     if (res && res != SG_LIB_CAT_RECOVERED)
@@ -287,14 +267,14 @@ do_safte_usage_statistics(int sg_fd, int do_hex, int do_raw, int verbose)
     int res;
     unsigned int rb_len;
     unsigned char *rb_buff;
-    unsigned long minutes;
+    unsigned int minutes;
 
     rb_len = 16 + safte_cfg.vendor_specific;
     rb_buff = (unsigned char *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=2 "
-                "to read usage statistics\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=2 to read "
+                "usage statistics\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 2, 0,
                             rb_buff, rb_len, 0, verbose);
     if (res) {
@@ -317,12 +297,10 @@ do_safte_usage_statistics(int sg_fd, int do_hex, int do_raw, int verbose)
         return 0;
     }
     printf("Usage Statistics:\n");
-    minutes = (rb_buff[0] << 24) + (rb_buff[1] << 16) +
-        (rb_buff[2] <<  8) + rb_buff[3];
-    printf("\tPower on Minutes: %ld\n", minutes);
-    minutes = (rb_buff[4] << 24) + (rb_buff[5] << 16) +
-        (rb_buff[6] <<  8) + rb_buff[7];
-    printf("\tPower on Cycles: %ld\n", minutes);
+    minutes = sg_get_unaligned_be32(rb_buff + 0);
+    printf("\tPower on Minutes: %u\n", minutes);
+    minutes = sg_get_unaligned_be32(rb_buff + 4);
+    printf("\tPower on Cycles: %u\n", minutes);
 
     free(rb_buff);
     return 0;
@@ -340,8 +318,8 @@ do_safte_slot_insertions(int sg_fd, int do_hex, int do_raw, int verbose)
     rb_buff = (unsigned char *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=3 "
-                "to read device insertions\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=3 to read "
+                "device insertions\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 3, 0,
                             rb_buff, rb_len, 0, verbose);
     if (res ) {
@@ -365,7 +343,7 @@ do_safte_slot_insertions(int sg_fd, int do_hex, int do_raw, int verbose)
     }
     printf("Slot insertions:\n");
     for (i = 0; i < safte_cfg.slots; i++) {
-        slot_status = (rb_buff[i * 2] << 8) + rb_buff[i * 2];
+        slot_status = sg_get_unaligned_be16(rb_buff + (i * 2));
         printf("\tSlot %d: %d insertions", i, slot_status);
     }
     free(rb_buff);
@@ -384,8 +362,8 @@ do_safte_slot_status(int sg_fd, int do_hex, int do_raw, int verbose)
     rb_buff = (unsigned char *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=4 "
-                "to read device slot status\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=4 to read "
+                "device slot status\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 4, 0,
                             rb_buff, rb_len, 0, verbose);
     if (res && res != SG_LIB_CAT_RECOVERED) {
@@ -433,8 +411,8 @@ do_safte_global_flags(int sg_fd, int do_hex, int do_raw, int verbose)
     rb_buff = (unsigned char *)malloc(rb_len);
 
     if (verbose > 1)
-        fprintf(stderr, "Use READ BUFFER,mode=vendor_specific,buff_id=5 "
-                "to read global flags\n");
+        pr2serr("Use READ BUFFER,mode=vendor_specific,buff_id=5 to read "
+                "global flags\n");
     res = sg_ll_read_buffer(sg_fd, RWB_MODE_VENDOR, 5, 0,
                             rb_buff, rb_len, 0, verbose);
     if (res ) {
@@ -489,8 +467,7 @@ do_safte_global_flags(int sg_fd, int do_hex, int do_raw, int verbose)
 static
 void usage()
 {
-    fprintf(stderr,
-            "Usage:  sg_safte [--config] [--devstatus] [--encstatus] "
+    pr2serr("Usage:  sg_safte [--config] [--devstatus] [--encstatus] "
             "[--flags] [--help]\n"
             "                 [--hex] [--insertions] [--raw] [--usage] "
             "[--verbose]\n"
@@ -546,6 +523,7 @@ main(int argc, char * argv[])
     int do_insertions = 0;
     const char * cp;
     char buff[48];
+    char b[80];
     struct sg_simple_inquiry_resp inq_resp;
     const char op_name[] = "READ BUFFER";
 
@@ -591,10 +569,10 @@ main(int argc, char * argv[])
                 ++verbose;
                 break;
             case 'V':
-                fprintf(stderr, "Version string: %s\n", version_str);
+                pr2serr("Version string: %s\n", version_str);
                 exit(0);
             default:
-                fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+                pr2serr("unrecognised option code 0x%x ??\n", c);
                 usage();
                 return SG_LIB_SYNTAX_ERROR;
         }
@@ -606,18 +584,23 @@ main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
     }
 
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
+    if (do_raw) {
+        if (sg_set_binary_mode(STDOUT_FILENO) < 0) {
+            perror("sg_set_binary_mode");
+            return SG_LIB_FILE_ERROR;
+        }
+    }
 
     if ((sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose)) < 0) {
         snprintf(ebuff, EBUFF_SZ, "sg_safte: error opening file: %s (rw)",
@@ -638,8 +621,8 @@ main(int argc, char * argv[])
             else
                 printf("  Peripheral device type: 0x%x\n", peri_type);
         } else {
-            fprintf(stderr, "sg_safte: %s doesn't respond to a SCSI "
-                    "INQUIRY\n", device_name);
+            pr2serr("sg_safte: %s doesn't respond to a SCSI INQUIRY\n",
+                    device_name);
             return SG_LIB_CAT_OTHER;
         }
     }
@@ -734,29 +717,15 @@ err_out:
     case 0:
     case SG_LIB_CAT_RECOVERED:
         break;
-    case SG_LIB_CAT_ABORTED_COMMAND:
-        fprintf(stderr, "%s: aborted command\n", op_name);
-        break;
-    case SG_LIB_CAT_NOT_READY:
-        fprintf(stderr, "%s: device not ready\n", op_name);
-        break;
-    case SG_LIB_CAT_UNIT_ATTENTION:
-        fprintf(stderr, "%s: unit attention\n", op_name);
-        break;
-    case SG_LIB_CAT_INVALID_OP:
-        fprintf(stderr, "%s: operation not supported\n", op_name);
-        break;
-    case SG_LIB_CAT_ILLEGAL_REQ:
-        fprintf(stderr, "%s: bad field in cdb\n", op_name);
-        break;
     default:
-        fprintf(stderr, "%s failed\n", op_name);
+        sg_get_category_sense_str(res, sizeof(b), b, verbose);
+        pr2serr("%s failed: %s\n", op_name, b);
         break;
     }
     ret = res;
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
             return SG_LIB_FILE_ERROR;
     }
