@@ -1,30 +1,8 @@
 /*
- * Copyright (c) 2004-2007 Hannes Reinecke, Christophe Varoqui and Douglas Gilbert.
+* Copyright (c) 2004-2015 Hannes Reinecke, Christophe Varoqui, Douglas Gilbert
  * All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- *
+ * Use of this source code is governed by a BSD-style
+ * license that can be found in the BSD_LICENSE file.
  */
 
 #include <unistd.h>
@@ -38,15 +16,17 @@
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
+#include "sg_unaligned.h"
+#include "sg_pr2serr.h"
 
-/* A utility program for the Linux OS SCSI subsystem.
+/* A utility program originally written for the Linux OS SCSI subsystem.
  *
  *
  * This program issues the SCSI command SET TARGET PORT GROUPS
  * to the given SCSI device.
  */
 
-static char * version_str = "1.2 20070926";
+static const char * version_str = "1.10 20151219";
 
 #define TGT_GRP_BUFF_LEN 1024
 #define MX_ALLOC_LEN (0xc000 + 0x80)
@@ -58,6 +38,28 @@ static char * version_str = "1.2 20070926";
 #define TPGS_STATE_OFFLINE 0xe          /* SPC-4 rev 9 */
 #define TPGS_STATE_TRANSITIONING 0xf
 
+/* See also table 306 - Target port group descriptor format in SPC-4 rev 36e */
+#ifndef __cplusplus
+
+static const unsigned char state_sup_mask[] = {
+        [TPGS_STATE_OPTIMIZED]     = 0x01,
+        [TPGS_STATE_NONOPTIMIZED]  = 0x02,
+        [TPGS_STATE_STANDBY]       = 0x04,
+        [TPGS_STATE_UNAVAILABLE]   = 0x08,
+        [TPGS_STATE_OFFLINE]       = 0x40,
+        [TPGS_STATE_TRANSITIONING] = 0x80,
+};
+
+#else
+
+// C++ does not support designated initializers
+static const unsigned char state_sup_mask[] = {
+    0x1, 0x2, 0x4, 0x8, 0x0, 0x0, 0x0, 0x0,
+    0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x80,
+};
+
+#endif  /* C or C++ ? */
+
 #define VPD_DEVICE_ID  0x83
 #define DEF_VPD_DEVICE_ID_LEN  252
 
@@ -88,39 +90,36 @@ static struct option long_options[] = {
 static void
 usage()
 {
-    fprintf(stderr, "Usage: "
-          "sg_stpg   [--active] [--help] [--hex] [--offline] [--optimized] "
-          "[--raw]\n"
-          "                 [--standby] [--state=S,S...] [--tp=P,P...] "
-          "[--unavailable]\n"
-          "                 [--verbose] [--version] DEVICE\n"
-          "  where:\n"
-          "    --active|-a        set asymm. access state to "
-          "active/non-optimized\n"
-          "    --help|-h          print out usage message\n"
-          "    --hex|-H           print out report response in hex, then "
-          "exit\n"
-          "    --offline|-l       set asymm. access state to offline, takes "
-          "relative\n"
-          "                       target port id, rather than target port "
-          "group id\n"
-          "    --optimized|-o     set asymm. access state to "
-          "active/optimized\n"
-          "    --raw|-r           output report response in binary to "
-          "stdout, then exit\n"
-          "    --standby|-s       set asymm. access state to standby\n"
-          "    --state=S,S.. |-S S,S...     list of states (values or "
-          "acronyms)\n"
-          "    --tp=P,P.. |-t P,P...        list of target port group "
-          "identifiers,\n"
-          "                                 or relative target port "
-          "identifiers\n"
-          "    --unavailable|-u   set asymm. access state to unavailable\n"
-          "    --verbose|-v       increase verbosity\n"
-          "    --version|-V       print version string and exit\n\n"
-          "Performs a SCSI SET TARGET PORT GROUPS command\n"
-          );
-
+    pr2serr("Usage: sg_stpg   [--active] [--help] [--hex] [--offline] "
+            "[--optimized] [--raw]\n"
+            "                 [--standby] [--state=S,S...] [--tp=P,P...] "
+            "[--unavailable]\n"
+            "                 [--verbose] [--version] DEVICE\n"
+            "  where:\n"
+            "    --active|-a        set asymm. access state to "
+            "active/non-optimized\n"
+            "    --help|-h          print out usage message\n"
+            "    --hex|-H           print out report response in hex, then "
+            "exit\n"
+            "    --offline|-l       set asymm. access state to offline, takes "
+            "relative\n"
+            "                       target port id, rather than target port "
+            "group id\n"
+            "    --optimized|-o     set asymm. access state to "
+            "active/optimized\n"
+            "    --raw|-r           output report response in binary to "
+            "stdout, then exit\n"
+            "    --standby|-s       set asymm. access state to standby\n"
+            "    --state=S,S.. |-S S,S...     list of states (values or "
+            "acronyms)\n"
+            "    --tp=P,P.. |-t P,P...        list of target port group "
+            "identifiers,\n"
+            "                                 or relative target port "
+            "identifiers\n"
+            "    --unavailable|-u   set asymm. access state to unavailable\n"
+            "    --verbose|-v       increase verbosity\n"
+            "    --version|-V       print version string and exit\n\n"
+            "Performs a SCSI SET TARGET PORT GROUPS command\n");
 }
 
 static void
@@ -135,7 +134,7 @@ dStrRaw(const char* str, int len)
 static int
 decode_target_port(unsigned char * buff, int len, int *d_id, int *d_tpg)
 {
-    int c_set, piv, assoc, desig_type, i_len;
+    int c_set, assoc, desig_type, i_len;
     int off, u;
     const unsigned char * ucp;
     const unsigned char * ip;
@@ -147,40 +146,40 @@ decode_target_port(unsigned char * buff, int len, int *d_id, int *d_tpg)
         ucp = buff + off;
         i_len = ucp[3];
         if ((off + i_len + 4) > len) {
-            fprintf(stderr, "    VPD page error: designator length longer "
-                    "than\n     remaining response length=%d\n", (len - off));
+            pr2serr("    VPD page error: designator length longer than\n     "
+                    "remaining response length=%d\n", (len - off));
             return SG_LIB_CAT_MALFORMED;
         }
         ip = ucp + 4;
         c_set = (ucp[0] & 0xf);
-        piv = ((ucp[1] & 0x80) ? 1 : 0);
+        /* piv = ((ucp[1] & 0x80) ? 1 : 0); */
         assoc = ((ucp[1] >> 4) & 0x3);
         desig_type = (ucp[1] & 0xf);
         switch (desig_type) {
         case 4: /* Relative target port */
             if ((1 != c_set) || (1 != assoc) || (4 != i_len)) {
-                fprintf(stderr, "      << expected binary code_set, target "
-                        "port association, length 4>>\n");
-                dStrHex((const char *)ip, i_len, 0);
+                pr2serr("      << expected binary code_set, target port "
+                        "association, length 4>>\n");
+                dStrHexErr((const char *)ip, i_len, 0);
                 break;
             }
-            *d_id = ((ip[2] << 8) | ip[3]);
+            *d_id = sg_get_unaligned_be16(ip + 2);
             break;
         case 5: /* (primary) Target port group */
             if ((1 != c_set) || (1 != assoc) || (4 != i_len)) {
-                fprintf(stderr, "      << expected binary code_set, target "
-                        "port association, length 4>>\n");
-                dStrHex((const char *)ip, i_len, 0);
+                pr2serr("      << expected binary code_set, target port "
+                        "association, length 4>>\n");
+                dStrHexErr((const char *)ip, i_len, 0);
                 break;
             }
-            *d_tpg = ((ip[2] << 8) | ip[3]);
+            *d_tpg = sg_get_unaligned_be16(ip + 2);
             break;
         default:
             break;
         }
     }
     if (-1 == *d_id || -1 == *d_tpg) {
-        fprintf(stderr, "VPD page error: no target port group information\n");
+        pr2serr("VPD page error: no target port group information\n");
         return SG_LIB_CAT_MALFORMED;
     }
     return 0;
@@ -229,7 +228,7 @@ transition_tpgs_states(struct tgtgrp *tgtState, int numgrp, int portgroup,
           return 1;
      }
 
-     if (!( (1 << newstate) & tgtState[i].valid )) {
+     if (!( state_sup_mask[newstate] & tgtState[i].valid )) {
           printf("Portgroup 0x%02x: Invalid state 0x%x\n",
                  portgroup, newstate);
           return 1;
@@ -273,10 +272,9 @@ encode_tpgs_states(unsigned char *buff, struct tgtgrp *tgtState, int numgrp)
 
      for (i = 0, desc = buff + 4; i < numgrp; desc += 4, i++) {
           desc[0] = tgtState[i].current & 0x0f;
-          desc[2] = (tgtState[i].id >> 8) & 0x0f;
-          desc[3] = tgtState[i].id & 0x0f;
+          sg_put_unaligned_be16((uint16_t)tgtState[i].id, desc + 2);
      }
-} 
+}
 
 /* Read numbers (up to 32 bits in size) from command line (comma separated */
 /* list). Assumed decimal unless prefixed by '0x', '0X' or contains */
@@ -299,26 +297,26 @@ build_port_arr(const char * inp, int * port_arr, int * port_arr_len,
         *port_arr_len = 0;
     k = strspn(inp, "0123456789aAbBcCdDeEfFhHxX,");
     if (in_len != k) {
-        fprintf(stderr, "build_port_arr: error at pos %d\n", k + 1);
+        pr2serr("build_port_arr: error at pos %d\n", k + 1);
         return 1;
     }
     for (k = 0; k < max_arr_len; ++k) {
         v = sg_get_num_nomult(lcp);
         if (-1 != v) {
             port_arr[k] = v;
-            cp = strchr(lcp, ',');
+            cp = (char *)strchr(lcp, ',');
             if (NULL == cp)
                 break;
             lcp = cp + 1;
         } else {
-            fprintf(stderr, "build_port_arr: error at pos %d\n",
+            pr2serr("build_port_arr: error at pos %d\n",
                     (int)(lcp - inp + 1));
             return 1;
         }
     }
     *port_arr_len = k + 1;
     if (k == max_arr_len) {
-        fprintf(stderr, "build_port_arr: array length exceeded\n");
+        pr2serr("build_port_arr: array length exceeded\n");
         return 1;
     }
     return 0;
@@ -347,7 +345,7 @@ build_state_arr(const char * inp, int * state_arr, int * state_arr_len,
         *state_arr_len = 0;
     k = strspn(inp, "0123456789aAbBcCdDeEfFhHnNoOsSuUxX,");
     if (in_len != k) {
-        fprintf(stderr, "build_state_arr: error at pos %d\n", k + 1);
+        pr2serr("build_state_arr: error at pos %d\n", k + 1);
         return 1;
     }
     for (k = 0; k < max_arr_len; ++k) {
@@ -356,9 +354,9 @@ build_state_arr(const char * inp, int * state_arr, int * state_arr_len,
             try_num = 0;
             switch (toupper(*lcp)) {
             case 'A':
-                if ('N' == toupper(*(lcp + 1))) 
+                if ('N' == toupper(*(lcp + 1)))
                     state_arr[k] = 1;
-                else if ('O' == toupper(*(lcp + 1))) 
+                else if ('O' == toupper(*(lcp + 1)))
                     state_arr[k] = 0;
                 else
                     try_num = 1;
@@ -373,8 +371,8 @@ build_state_arr(const char * inp, int * state_arr, int * state_arr_len,
                 state_arr[k] = 3;
                 break;
             default:
-                fprintf(stderr, "build_state_arr: expected 'ao', 'an', 'o', "
-                        "'s' or 'u' at pos %d\n", (int)(lcp - inp + 1));
+                pr2serr("build_state_arr: expected 'ao', 'an', 'o', 's' or "
+                        "'u' at pos %d\n", (int)(lcp - inp + 1));
                 return 1;
             }
         }
@@ -383,22 +381,22 @@ build_state_arr(const char * inp, int * state_arr, int * state_arr_len,
             if (((v >= 0) && (v <= 3)) || (14 ==v))
                 state_arr[k] = v;
             else if (-1 == v) {
-                fprintf(stderr, "build_state_arr: error at pos %d\n",
+                pr2serr("build_state_arr: error at pos %d\n",
                         (int)(lcp - inp + 1));
                 return 1;
             } else {
-                fprintf(stderr, "build_state_arr: expect 0,1,2,3 or 14\n");
+                pr2serr("build_state_arr: expect 0,1,2,3 or 14\n");
                 return 1;
             }
         }
-        cp = strchr(lcp, ',');
+        cp = (char *)strchr(lcp, ',');
         if (NULL == cp)
             break;
         lcp = cp + 1;
     }
     *state_arr_len = k + 1;
     if (k == max_arr_len) {
-        fprintf(stderr, "build_state_arr: array length exceeded\n");
+        pr2serr("build_state_arr: array length exceeded\n");
         return 1;
     }
     return 0;
@@ -408,7 +406,7 @@ build_state_arr(const char * inp, int * state_arr, int * state_arr_len,
 int
 main(int argc, char * argv[])
 {
-    int sg_fd, k, off, res, c, report_len, tgt_port_count, trunc;
+    int sg_fd, k, off, res, c, report_len, tgt_port_count;
     unsigned char reportTgtGrpBuff[TGT_GRP_BUFF_LEN];
     unsigned char setTgtGrpBuff[TGT_GRP_BUFF_LEN];
     unsigned char rsp_buff[MX_ALLOC_LEN + 2];
@@ -423,6 +421,7 @@ main(int argc, char * argv[])
     int port_arr[MAX_PORT_LIST_ARR_LEN];
     int port_arr_len = 0;
     int state_arr[MAX_PORT_LIST_ARR_LEN];
+    char b[80];
     int state_arr_len = 0;
     int portgroup = -1;
     int relport = -1;
@@ -475,10 +474,10 @@ main(int argc, char * argv[])
             ++verbose;
             break;
         case 'V':
-            fprintf(stderr, "Version: %s\n", version_str);
+            pr2serr("Version: %s\n", version_str);
             return 0;
         default:
-            fprintf(stderr, "unrecognised option code 0x%x ??\n", c);
+            pr2serr("unrecognised option code 0x%x ??\n", c);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -490,8 +489,7 @@ main(int argc, char * argv[])
         }
         if (optind < argc) {
             for (; optind < argc; ++optind)
-                fprintf(stderr, "Unexpected extra argument: %s\n",
-                        argv[optind]);
+                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -512,8 +510,8 @@ main(int argc, char * argv[])
         }
     }
     if ((state >= 0) && (state_arr_len > 0)) {
-        fprintf(stderr, "either use individual state option or '--state=' "
-                "but not both\n");
+        pr2serr("either use individual state option or '--state=' but not "
+                "both\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
@@ -524,14 +522,13 @@ main(int argc, char * argv[])
         state_arr_len = 0;
     }
     if (state_arr_len > port_arr_len) {
-        fprintf(stderr, "'state=' list longer than expected\n");
+        pr2serr("'state=' list longer than expected\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     if ((port_arr_len > 0) && (0 == state_arr_len)) {
         if (-1 == state) {
-            fprintf(stderr, "target port list given but no state "
-                    "indicated\n");
+            pr2serr("target port list given but no state indicated\n");
             usage();
             return SG_LIB_SYNTAX_ERROR;
         }
@@ -545,20 +542,19 @@ main(int argc, char * argv[])
         state_arr_len = port_arr_len;
     }
     if (port_arr_len != state_arr_len) {
-        fprintf(stderr, "'state=' and '--tp=' lists mismatched\n");
+        pr2serr("'state=' and '--tp=' lists mismatched\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
 
     if (NULL == device_name) {
-        fprintf(stderr, "missing device name!\n");
+        pr2serr("missing device name!\n");
         usage();
         return SG_LIB_SYNTAX_ERROR;
     }
     sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
     if (sg_fd < 0) {
-        fprintf(stderr, "open error: %s: %s\n", device_name,
-                safe_strerror(-sg_fd));
+        pr2serr("open error: %s: %s\n", device_name, safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
     }
 
@@ -566,18 +562,18 @@ main(int argc, char * argv[])
         res = sg_ll_inquiry(sg_fd, 0, 1, VPD_DEVICE_ID, rsp_buff,
                             DEF_VPD_DEVICE_ID_LEN, 1, verbose);
         if (0 == res) {
-            report_len = ((rsp_buff[2] << 8) + rsp_buff[3]) + 4;
+            report_len = sg_get_unaligned_be16(rsp_buff + 2) + 4;
             if (VPD_DEVICE_ID != rsp_buff[1]) {
-                fprintf(stderr, "invalid VPD response; probably a STANDARD "
-                        "INQUIRY response\n");
+                pr2serr("invalid VPD response; probably a STANDARD INQUIRY "
+                        "response\n");
                 if (verbose) {
-                    fprintf(stderr, "First 32 bytes of bad response\n");
-                    dStrHex((const char *)rsp_buff, 32, 0);
+                    pr2serr("First 32 bytes of bad response\n");
+                    dStrHexErr((const char *)rsp_buff, 32, 0);
                 }
                 return SG_LIB_CAT_MALFORMED;
             }
             if (report_len > MX_ALLOC_LEN) {
-                fprintf(stderr, "response length too long: %d > %d\n", report_len,
+                pr2serr("response length too long: %d > %d\n", report_len,
                         MX_ALLOC_LEN);
                 return SG_LIB_CAT_MALFORMED;
             } else if (report_len > DEF_VPD_DEVICE_ID_LEN) {
@@ -585,26 +581,25 @@ main(int argc, char * argv[])
                                   report_len, 1, verbose))
                     return SG_LIB_CAT_OTHER;
             }
-            decode_target_port(rsp_buff + 4, report_len - 4, &relport, &portgroup);
+            decode_target_port(rsp_buff + 4, report_len - 4, &relport,
+                               &portgroup);
             printf("Device is at port Group 0x%02x, relative port 0x%02x\n",
                    portgroup, relport);
         }
 
         memset(reportTgtGrpBuff, 0x0, sizeof(reportTgtGrpBuff));
-        trunc = 0;
+        /* trunc = 0; */
 
-        res = sg_ll_report_tgt_prt_grp(sg_fd, reportTgtGrpBuff,
-                                sizeof(reportTgtGrpBuff), 1, verbose);
+        res = sg_ll_report_tgt_prt_grp2(sg_fd, reportTgtGrpBuff,
+                                        sizeof(reportTgtGrpBuff), 0, 1,
+                                        verbose);
         ret = res;
         if (0 == res) {
-            report_len = (reportTgtGrpBuff[0] << 24) +
-                         (reportTgtGrpBuff[1] << 16) + 
-                         (reportTgtGrpBuff[2] << 8) +
-                         reportTgtGrpBuff[3] + 4;
+            report_len = sg_get_unaligned_be32(reportTgtGrpBuff + 0) + 4;
             if (report_len > (int)sizeof(reportTgtGrpBuff)) {
-                trunc = 1;
-                fprintf(stderr, "  <<report too long for internal buffer,"
-                        " output truncated\n");
+                /* trunc = 1; */
+                pr2serr("  <<report too long for internal buffer, output "
+                        "truncated\n");
                 report_len = (int)sizeof(reportTgtGrpBuff);
             }
             if (raw) {
@@ -615,7 +610,7 @@ main(int argc, char * argv[])
                 printf("Report list length = %d\n", report_len);
             if (hex) {
                 if (verbose)
-                    fprintf(stderr, "\nOutput response in hex:\n");
+                    printf("\nOutput response in hex:\n");
                 dStrHex((const char *)reportTgtGrpBuff, report_len, 1);
                 goto err_out;
             }
@@ -626,32 +621,24 @@ main(int argc, char * argv[])
                  k += off, ucp += off, numgrp ++) {
 
                 printf("  target port group id : 0x%x , Pref=%d\n",
-                       (ucp[2] << 8) + ucp[3], !!(ucp[0] & 0x80));
+                       sg_get_unaligned_be16(ucp + 2), !!(ucp[0] & 0x80));
                 printf("    target port group asymmetric access state : ");
                 printf("0x%02x", ucp[0] & 0x0f);
                 printf("\n");
-                tgtStatePtr->id = (ucp[2] << 8) + ucp[3];
+                tgtStatePtr->id = sg_get_unaligned_be16(ucp + 2);
                 tgtStatePtr->current = ucp[0] & 0x0f;
                 tgtStatePtr->valid = ucp[1];
-            
+
                 tgt_port_count = ucp[7];
 
                 tgtStatePtr++;
                 off = 8 + tgt_port_count * 4;
             }
-        } else if (SG_LIB_CAT_INVALID_OP == res)
-            fprintf(stderr, "Report Target Port Groups command not supported\n");
-        else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-            fprintf(stderr, "bad field in Report Target Port Groups cdb "
-                    "including unsupported service action\n");
-        else if (SG_LIB_CAT_UNIT_ATTENTION == res)
-            fprintf(stderr, "Report Target Port Groups, unit attention\n");
-        else if (SG_LIB_CAT_ABORTED_COMMAND == res)
-            fprintf(stderr, "Report Target Port Groups, aborted command\n");
-        else {
-            fprintf(stderr, "Report Target Port Groups command failed\n");
+        } else {
+            sg_get_category_sense_str(res, sizeof(b), b, verbose);
+            pr2serr("Report Target Port Groups: %s\n", b);
             if (0 == verbose)
-                fprintf(stderr, "    try '-v' for more information\n");
+                pr2serr("    try '-v' for more information\n");
         }
         if (0 != res)
              goto err_out;
@@ -663,7 +650,7 @@ main(int argc, char * argv[])
         transition_tpgs_states(tgtGrpState, numgrp, portgroup, state);
 
         memset(setTgtGrpBuff, 0x0, sizeof(setTgtGrpBuff));
-        trunc = 0;
+        /* trunc = 0; */
 
         encode_tpgs_states(setTgtGrpBuff, tgtGrpState, numgrp);
         report_len = numgrp * 4 + 4;
@@ -671,8 +658,7 @@ main(int argc, char * argv[])
         memset(setTgtGrpBuff, 0x0, sizeof(setTgtGrpBuff));
         for (k = 0, ucp = setTgtGrpBuff + 4; k < port_arr_len; ++k, ucp +=4) {
             ucp[0] = state_arr[k] & 0xf;
-            ucp[2] = (port_arr[k] >> 8) & 0xff;
-            ucp[3] = port_arr[k] & 0xff;
+            sg_put_unaligned_be16((uint16_t)port_arr[k], ucp + 2);
         }
         report_len = port_arr_len * 4 + 4;
     }
@@ -681,25 +667,17 @@ main(int argc, char * argv[])
 
     if (0 == res)
         goto err_out;
-    else if (SG_LIB_CAT_INVALID_OP == res)
-        fprintf(stderr, "Set Target Port Groups command not supported\n");
-    else if (SG_LIB_CAT_ILLEGAL_REQ == res)
-        fprintf(stderr, "bad field in Report Target Port Groups cdb "
-                "including unsupported service action\n");
-    else if (SG_LIB_CAT_UNIT_ATTENTION == res)
-        fprintf(stderr, "Set Target Port Groups, unit attention\n");
-    else if (SG_LIB_CAT_ABORTED_COMMAND == res)
-        fprintf(stderr, "Set Target Port Groups, aborted command\n");
     else {
-        fprintf(stderr, "Set Target Port Groups command failed\n");
+        sg_get_category_sense_str(res, sizeof(b), b, verbose);
+        pr2serr("Set Target Port Groups: %s\n", b);
         if (0 == verbose)
-            fprintf(stderr, "    try '-v' for more information\n");
+            pr2serr("    try '-v' for more information\n");
     }
 
 err_out:
     res = sg_cmds_close_device(sg_fd);
     if (res < 0) {
-        fprintf(stderr, "close error: %s\n", safe_strerror(-res));
+        pr2serr("close error: %s\n", safe_strerror(-res));
         if (0 == ret)
             return SG_LIB_FILE_ERROR;
     }
